#!/bin/env python
# -*- coding: utf-8 -*-

# ------------------------------------------------------------
# environment.py
# 
# pick some current environment infos
#
# Autor: Oliver Maurhart, <oliver.maurhart@ait.ac.at>
#
# Copyright (C) 2013 AIT Austrian Institute of Technology
# AIT Austrian Institute of Technology GmbH
# Donau-City-Strasse 1 | 1220 Vienna | Austria
# http://www.ait.ac.at
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation version 2.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, 
# Boston, MA  02110-1301, USA.
# ------------------------------------------------------------


# ------------------------------------------------------------
# imports

import os
import os.path
import sys


# ------------------------------------------------------------
# code


class Environment(object):
    
    """Hold some nifty environment stuff in a dedicated class."""
    
    def __init__(self, application = None):
        
        # if we ain't got a path to start from, all is valid/lost
        if len(sys.path[0]) == 0:
            self._prefix_path = ''
            self._data_path = ''
            return
        
        # the prefix path
        #
        # - on Linux: this is ../../ to the current executable
        #   e.g. "/usr/bin/myprogram" --> "/usr"
        #
        # - on Windows (inkl. Cygwin): this is the installation folder
        #   e.g. "C:/Program Files/MyProgram/myprogam" --> "C:/Program Files/MyProgram"
        #
        if sys.platform == 'linux2':
            self._prefix_path = os.path.split(sys.path[0])[0]
        elif sys.platform == 'win32' or sys.platform == 'cygwin':
            self._prefix_path = sys.path[0]
            
        # the data path where all data files are stored
        if sys.platform == 'linux2':
            if not application is None:
                self._data_path = os.path.join(self._prefix_path, os.path.join('share', application))
            else:
                self._data_path = os.path.join(self._prefix_path, 'share')
        elif sys.platform == 'win32' or sys.platform == 'cygwin':
            self._data_path = self._prefix_path
        else:
            raise OSError()

            
    def data_path_get(self):
        """dat_path get"""
        return self._data_path
        
    data_path = property(data_path_get)
            
            
    def prefix_path_get(self):
        """prefix_path get"""
        return self._prefix_path
        
    prefix_path = property(prefix_path_get)
    
    
    def log_path_get(self):
        
        """the path where log files should be stored"""
        
        user_log_path = os.path.expanduser(os.path.join('~', '.log'))
        
        if sys.platform == 'linux2':
            
            if os.access('/var/log', os.W_OK):
                return '/var/log'
            
            print('no permissions to write log files in /var/log, switching to ~/.log')
            
            if not os.path.exists(user_log_path):
                os.mkdir(user_log_path)
            elif not os.path.isdir(user_log_path):
                raise IOError(user_log_path + ': not a folder')
            
            return user_log_path
        
        elif sys.platform == 'win32' or sys.platform == 'cygwin':

            if not os.path.exists(user_log_path):
                os.mkdir(user_log_path)
            elif not os.path.isdir(user_log_path):
                raise IOError(user_log_path + ': not a folder')

            return user_log_path
        
        else:
            raise OSError()
            
    log_path = property(log_path_get)
            
            
# test method
def test():

    """Test: class Environment"""
    e = Environment('My Application')
    print('prefix_path: "{0}"'.format(e.prefix_path))
    print('  data_path: "{0}"'.format(e.data_path))
    print('   log_path: "{0}"'.format(e.log_path))


# test the module			
if __name__ == '__main__':
    test()
